//Parent class for the collimators

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <cmath>
#include "Collimator.h"
using namespace std;

Collimator::Collimator(const double& ALFX, const double& ALFY, const double& APER_1, const double& APER_2, const double& APER_3, const double& APER_4, const string& APERTYPE, const double& BETX, const double& BETY, const double& DPX, const double& DPY, const double& DX, const double& DY, const string& KEYWORD, const double& L, const double& MUX, const double& MUY, const string& NAME, const double& PTC, const double& PXC, const double& PYC, const double& S, const double& TC, const double& XC, const double& YC, const double& K0L, const double& K0SL, const double& K1L, const double& K1SL, const double& K2L, const double& K2SL, const string& PARENT, const string& meth, const long double& hgap, const long double& hgap2, const double& collang, const long double& pdepth, const long double& pdepth2, const double& tcang, const double& nsig)
    : Element(ALFX, ALFY, APER_1, APER_2, APER_3, APER_4, APERTYPE, BETX, BETY, DPX, DPY, DX, DY, KEYWORD, L, MUX, MUY, NAME, PTC, PXC, PYC, S, TC, XC, YC, K0L, K0SL, K1L, K1SL, K2L, K2SL, PARENT),
      method(meth),
      hgap(hgap),
      hgap2(hgap2),
      phi(collang),
      pdepth(pdepth),
      pdepth2(pdepth2),
      tcang(tcang),
      nsig(nsig)
{};

Collimator::Collimator(const double& ALFX, const double& ALFY, const double& APER_1, const double& APER_2, const double& APER_3, const double& APER_4, const string& APERTYPE, const double& BETX, const double& BETY, const double& DPX, const double& DPY, const double& DX, const double& DY, const string& KEYWORD, const double& L, const double& MUX, const double& MUY, const string& NAME, const double& PTC, const double& PXC, const double& PYC, const double& S, const double& TC, const double& XC, const double& YC, const double& K0L, const double& K0SL, const double& K1L, const double& K1SL, const double& K2L, const double& K2SL, const string& PARENT, const string& meth, const long double& hgap, const long double& hgap2, const double& collang, const long double& pdepth, const long double& pdepth2, const double& tcang, const double& nsig, const long double& Bmax, const long double& thicknessMagneticField, const double& energyPerIon, const double& mass)
    : Element(ALFX, ALFY, APER_1, APER_2, APER_3, APER_4, APERTYPE, BETX, BETY, DPX, DPY, DX, DY, KEYWORD, L, MUX, MUY, NAME, PTC, PXC, PYC, S, TC, XC, YC, K0L, K0SL, K1L, K1SL, K2L, K2SL, PARENT),
      method(meth),
      hgap(hgap),
      hgap2(hgap2),
      phi(collang),
      pdepth(pdepth),
      pdepth2(pdepth2),
      tcang(tcang),
      nsig(nsig),
      Bmax(Bmax),
      thicknessMagneticField(thicknessMagneticField),
      energyPerIon(energyPerIon),
      mass(mass)
{};

Collimator::Collimator(Element elt, const double& tcang, const double& nsig, const string& meth)
    : Element(elt),
      method(meth),
      tcang(tcang),
      nsig(nsig)
{};

Collimator::Collimator(Element elt, const double& tcang, const double& nsig, const string& meth, const string& material)
    : Element(elt),
      method(meth),
      tcang(tcang),
      nsig(nsig),
      material(material)
{};

Collimator::Collimator(const Collimator& obj)
    : Element(obj),
      method(obj.method),
      hgap(obj.hgap),
      hgap2(obj.hgap2),
      phi(obj.phi),
      pdepth(obj.pdepth),
      pdepth2(obj.pdepth2),
      tcang(obj.tcang),
      nsig(obj.nsig),
      material(obj.material),
      crossSectionPath(obj.crossSectionPath),
      Bmax(obj.Bmax),
      thicknessMagneticField(obj.thicknessMagneticField),
      energyPerIon(obj.energyPerIon),
      mass(obj.mass)
{}


void Collimator::affiche()
{

    this->Element::affiche();
    cout << "Method: " << method << endl;

};

void Collimator::collipassInteraction(Particle& p1, double Apr, double Zpr, double betgam, double lcoll)
{

    double Navo(6.022137e23);
    double s(0);
    double atarget, rho, stre;
    string abbreviation;

    ifstream entree;
    string filename;
    filename = this->crossSectionPath + "materialinfo.csv";;
    entree.open(filename.c_str());

    if (entree.fail()) {
        cerr << "Error: impossible to read the file " << filename << endl;
    } else {

        string word;
        double nbre;

        getline(entree, word);

        while (!entree.eof()) {

            entree >> ws;

            getline(entree, word, ',');

            if (word == "atarget") {
                entree >> atarget;//mass number of collimator
            } else if (word == "rho") {
                entree >> rho;//density of collimator material
            } else if (word == "stre") {
                entree >> stre;//total EM cross-section
            } else if (word == "abbreviation") {
                entree >> abbreviation;//chemical symbol for collimator material
            } else {
                entree >> word;
            }

        }//end while(!entree.eof())

    }
    entree.close();

    double Atarget;

    Atarget = atarget * 0.001;


    while (s < lcoll) { //loop until particle is lost or exits the collimator

        string file;
        ifstream entree1;
        double sn, wx, wy, dpopdx;

        std::stringstream apzp;
        apzp << p1.Ap0 << p1.Zp0;
        file = this->crossSectionPath + abbreviation + apzp.str() + "nuclEM.dat";
        entree1.open(file.c_str());

        if (entree1.fail()) {
            cerr << "Warning!! The file " << file << " does not have a valid fomat or does not exist." << endl;

            //we calculate analytical approximations to the EMD and hadronic cross section using a BCV parametrization of an impact-parameter cutoff model

            double stee, Bbcv, steh, ste, intlengthe;
            stee = stre * p1.Zp0 * (p1.Ap0 - p1.Zp0) / p1.Ap0 / (Zpr * (Apr - Zpr) / Apr); //approximation of electromagnetic dissociation cross section
            Bbcv = 1.34 * (pow(p1.Ap0, 1. / 3.) + pow(atarget, 1. / 3.) - 0.75 * (pow(p1.Ap0, -1. / 3.) + pow(atarget, -1. / 3.))); //radius of the overlap between two "balls"
            Bbcv = Bbcv / 1e15;
            steh = M_PI * Bbcv * Bbcv * 1e31; //area of overlap = hadronic cross section
            ste = steh + stee;//total cross section
            intlengthe = Atarget / (Navo * rho * ste * 1e-31); //approximation to the interaction length

            sn = s - intlengthe * log(-p1.random()); //distance between the beginning of the collimator and the next interaction

            if ((sn > lcoll) && (p1.Ap0 > 1) && (p1.Zp0 <= p1.Ap0)) { //particle is going out of the collimator
                BeBloMuSca(wx, wy, dpopdx, betgam, filename, rho, lcoll - s, atarget, p1); //calculating change in dx/ds, dy/ds and momentum
                p1.coordonnees[1][0] = p1.coordonnees[0][0];//x coordinate when particle hits
                p1.coordonnees[1][1] = p1.coordonnees[0][1] + wx;//xs=dx/ds when particle hits
                p1.coordonnees[1][2] = p1.coordonnees[0][2];//y coordinate when particle hits
                p1.coordonnees[1][3] = p1.coordonnees[0][3] + wy;//ys=dy/ds when particle hits
                p1.coordonnees[1][4] = p1.coordonnees[0][4] - dpopdx;//dpop = momentum when the particle hits
                return;
            } else {//particle is absorbed
                p1.Ap0 = 0;
                p1.Zp0 = -1;
                p1.coordonnees[1][0] = 0;
                p1.coordonnees[1][1] = 0;
                p1.coordonnees[1][2] = 0;
                p1.coordonnees[1][3] = 0;
                p1.coordonnees[1][4] = 0;
                return;
            }

        } else {

            double intlength, sigt;
            vector <double> da, dz, isig;

            genipsfastnewxc(sigt, isig, da, dz, file, p1.Ap0, p1.Zp0);

            if (da.size() == 0) {
                return;
            }

            intlength = Atarget / (Navo * rho * sigt * 1e-31);

            sn = s - intlength * log(-p1.random()); //Sample from exponential distribution. sn is the position for the next interaction

            if (sn > lcoll) { //particle exits collimator
                BeBloMuSca(wx, wy, dpopdx, betgam, filename, rho, lcoll - s, atarget, p1);
                p1.coordonnees[1][0] = p1.coordonnees[0][0];
                p1.coordonnees[1][1] = p1.coordonnees[0][1] + wx;
                p1.coordonnees[1][2] = p1.coordonnees[0][2];
                p1.coordonnees[1][3] = p1.coordonnees[0][3] + wy;
                p1.coordonnees[1][4] = p1.coordonnees[0][4] - dpopdx;
                return;
            } else if (lcoll - sn > 10 * intlength) { //particle is lost if it's moving more than 10 interaction lengths inside the collimator
                p1.Ap0 = 0;
                p1.Zp0 = -3;
                p1.coordonnees[1][0] = 0;
                p1.coordonnees[1][1] = 0;
                p1.coordonnees[1][2] = 0;
                p1.coordonnees[1][3] = 0;
                p1.coordonnees[1][4] = 0;
                return;
            }

            double dao, dzo, r;
            int chan(0);

            r = -p1.random();
            int temp(0);

            //deciding which new isotope is created
            for (int i(0); i < isig.size(); ++i) {
                if (r > isig[i]) {
                    temp = temp + 1;
                } else {
                    chan = temp;
                }
            }

            dao = da[chan - 1]; //decrease in mass number
            dzo = dz[chan - 1]; //decrease in charge

            if (dao == -999) { //corresponding to the omitted reactions, see genipsfastnewc. Particle assumed lost
                p1.Ap0 = 0;
                p1.Zp0 = -2;
                p1.coordonnees[1][0] = 0;
                p1.coordonnees[1][1] = 0;
                p1.coordonnees[1][2] = 0;
                p1.coordonnees[1][3] = 0;
                p1.coordonnees[1][4] = 0;
                return;
            }

            BeBloMuSca(wx, wy, dpopdx, betgam, filename, rho, sn - s, atarget, p1); //calculating change in dx/ds, dy/ds and momentum

            p1.coordonnees[0][1] = p1.coordonnees[0][1] + wx;
            p1.coordonnees[0][3] = p1.coordonnees[0][3] + wy;
            p1.coordonnees[0][4] = p1.coordonnees[0][4] - dpopdx;
            p1.Ap0 = p1.Ap0 - dao;
            p1.Zp0 = p1.Zp0 - dzo;

            s = sn;//'moving' particle to the place for the next interaction
        }
        entree1.close();
    }//end while(s < this->lcoll)

};


void Collimator::genipsfastnewxc(double& sigt, vector <double>& isig, vector <double>& da, vector <double>& dz, string path, double a, double z)
{

    ifstream entree;
    vector <double> temp1, temp2, temp3, sig, temp11, temp22, temp33;
    double maximum(0);

    entree.open(path.c_str());

    if (entree.fail()) {
        cerr << "We do not have cross-section information for the isotope with mass number " << a << " and charge " << z << " , i.e. the file " << path << "does not exist or has a wrong format."  << endl;
    } else {

        double nber1, nber2, nber3;
        vector <double> sig;


        entree >> nber1 >> nber2 >> nber3;

        sigt = nber3;

        while (!entree.eof()) {

            entree >> nber1 >> nber2 >> nber3;

            temp1.push_back(nber1);
            temp2.push_back(nber2);
            temp3.push_back(nber3);
        }
    }
    entree.close();

    double sum(0), msig;

    for (int j(0); j < temp1.size(); ++j) {
        if ((temp1[j] < a / 2) && (temp1[j] > 0)) { //only considering reactions where the mass decreases and the particle doesn't loose more than half its mass

            temp11.push_back(temp1[j]);
            temp22.push_back(temp2[j]);
            temp33.push_back(temp3[j]);
        }
    }

    for (int k(0); k < temp33.size(); ++k) {
        if (temp33[k] > maximum) {
            maximum = temp33[k];
        }
    }

    for (int j(0); j < temp33.size(); ++j) {
        if (temp33[j] > maximum / 100) { //excluding reactions with very small cross-section
            sig.push_back(temp33[j]);//cross sections
            da.push_back(temp11[j]);//change in mass number
            dz.push_back(temp22[j]);//change in charge
            sum = sum + temp33[j];
        }
    }
    temp1.clear();
    temp2.clear();
    temp3.clear();
    temp11.clear();
    temp22.clear();
    temp33.clear();

    if (sig.size() == 0) {
        cout << "No valid change in the cross-section." << endl;
        return;
    }
    msig = sigt - sum; //sum of the excluded cross sections

    sig.push_back(sig[sig.size() - 1]);
    da.push_back(da[da.size() - 1]);
    dz.push_back(dz[dz.size() - 1]);

    for (int k(sig.size() - 2); k > 0; --k) {
        sig[k] = sig[k - 1];
        da[k] = da[k - 1];
        dz[k] = dz[k - 1];
    }

    sig[0] = msig;
    da[0] = -999;
    da.push_back(-998);
    dz[0] = sigt;
    dz.push_back(-998);

    vector <double> hsign;

    for (int m(0); m < sig.size(); ++m) {
        hsign.push_back(sig[m] / sigt); //the fraction of the total cross section for each channel, with the missing cross sections as first element
    }

    //add up the fractional partial cross sections in a vector isig. First element is 0.
    //the different channels will then correspond to different intervals between 0 and 1 with lengths corresponding to their probability.

    isig.push_back(0);

    for (int p(1); p <= sig.size(); ++p) {
        isig.push_back(hsign[p - 1] + isig[p - 1]);
    }

};


void Collimator::BeBloMuSca(double& wx, double& wy, double& dpopdx, double betgam, string path, double rho, double L, double At, Particle p1)
{

    double me(510998.9);//electron mass [eV/c2]
    double K(0.307075e5);//K/A in table 23.1, PDG page 163
    double beta(sqrt(betgam * betgam / (betgam * betgam + 1))); //relativistic beta
    double gamma(betgam / beta); //relativistic gamma
    double X = log10(betgam);
    double Zt, X0, X1, m, a, C, M0, X0MS, Tmax, delta, wrms, Ii, pr, w, dEdx;

    ifstream entree;

    entree.open(path.c_str());

    if (entree.fail()) {
        cerr << "Warning: problem with the file '" << path << "'." << endl;
    } else {

        string word;
        double nbre;

        while (!entree.eof()) {

            getline(entree, word, ',');

            if (word == "ztarget") {
                getline(entree, word);
                Zt = atof(word.c_str());//atom number of collimator
            } else if (word == "X0") {
                getline(entree, word);
                X0 = atof(word.c_str());//Sternheimer parameter needed for density correction (not checked)
            } else if (word == "X1") {
                getline(entree, word);
                X1 = atof(word.c_str());//Sternheimer parameter needed for density correction (not checked)
            } else if (word == "m") {
                getline(entree, word);
                m = atof(word.c_str());//Sternheimer parameter needed for density correction (not checked)
            } else if (word == "a") {
                getline(entree, word);
                a = atof(word.c_str());//Sternheimer parameter needed for density correction (not checked)
            } else if (word == "C") {
                getline(entree, word);
                C = atof(word.c_str());//Sternheimer parameter needed for density correction (not checked)
            } else {
                getline(entree, word);
            }
        }//end while(!entree.eof())
        entree.close();
    }

    M0 = 931.494e6 * p1.Ap0; //mass of ion [MeV/c2]
    X0MS = 716.4 * At / (Zt * (Zt + 1) * log(287 / sqrt(Zt))) / (rho / 1000);
    if (Zt < 13) { //mean excitation potential [eV]
        Ii = Zt * (12 + 7 / Zt);
    } else {
        Ii = Zt * (9.76 + 58.8 * pow(Zt, -1.19));
    }

    Tmax = 2 * me * betgam * betgam / (1 + 2 * gamma * me / M0 + (me / M0) * (me / M0)); //Tmax (in eV) in Bethe-Bloch, PDG page 164

    //DENSITY CORRECTION

    if (X < X0) {
        delta = 0;//valid for non-conductors
    } else if ((X0 < X) && (X < X1)) {
        delta = 4.6052 * X + C + a * pow(X1 - X, m);
    } else {
        delta = 4.6052 * X + C;
    }

    L = L * 100;
    wrms = 19.2333e6 / (beta * betgam * M0) * p1.Zp0 * sqrt(L / X0MS) * (1 + 0.038 * log(L / X0MS)); //rms angular deviation from mult.scatt. Compare PDG page 166 - where does the prefactor come from? --from the fact that it's 3D not plane.

    L = L / 100;


    pr = -p1.random() * 2 * M_PI; //This line and the next samples from a 2D gaussian distribution the angular deviation in x and y.
    w = wrms * sqrt(-log(1 + p1.random()));

    //In order to make it analytically possible to sample, it is converted to polar coordinates, and the radius w and angle phi are sampled

    wx = w * cos(pr); //change in dx/ds
    wy = w * sin(pr); //change in dy/ds

    dEdx = rho * K * p1.Zp0 * p1.Zp0 * Zt / At / (beta * beta) * (0.5 * log(2 * me * betgam * betgam * Tmax / (Ii * Ii)) - beta * beta - delta / 2); //energy loss acc. to Bete-Bloch on page 163 in PDG [eV/Meter]

    dpopdx = L * dEdx / (M0 * gamma) / (beta * beta); //delta p/p caused by energy change due to Bethe-Bloch
};
