#ifndef ELEMENT_H
#define ELEMENT_H
#include <iostream>
#include <vector>
#include <string>
using namespace std;

/*
#################################################################################################################################

This class describes all elements that are in the accelerator, except the collimators, which are inheriting from this class.

#################################################################################################################################
*/


class Element
{


public:

    //===================================Constructors, destructor========================================

    Element() {};

    Element(const double& ALFX, const double& ALFY, const double& APER_1, const double& APER_2, const double& APER_3, const double& APER_4, const string& APERTYPE, const double& BETX, const double& BETY, const double& DPX, const double& DPY, const double& DX, const double& DY, const string& KEYWORD, const double& L, const double& MUX, const double& MUY, const string& NAME, const double& PTC, const double& PXC, const double& PYC, const double& S, const double& TC, const double& XC, const double& YC, const double& K0L, const double& K0SL, const double& K1L, const double& K1SL, const double& K2L, const double& K2SL, const string& PARENT);

    Element(const Element& obj);

    virtual ~Element() {};

    Element operator=(const Element& elt);


    //displays the parameters of the element

    virtual void affiche();


    double ALFX;//correlation function in x (ALFX = -0.5*BETX'
    double ALFY;//correlation function in y

    //the following terms are aperture parameters
    double APER_1;
    double APER_2;
    double APER_3;
    double APER_4;
    string APERTYPE;//type of aperture (RECTANGLE, RECTELLIPSE, CIRCLE, ELLIPSE or NONE)
    double BETX;//amplitude function in x [m]
    double BETY;//amplitude function in y [m]
    double DPX;//dispertion Dpx
    double DPY;//dispertion Dpy
    double DX;//dispertion in x [m]
    double DY;//dispertion in y [m]
    string KEYWORD;
    double L;//length of the element [m]
    double MUX;//phase advance in x [2*pi]
    double MUY;//phase advance in y [2*pi]
    string NAME;
    double PTC;//energy error, divided by the reference momentum times the velocity of light
    double PXC;//horizontal canonical momentum of the (closed) orbit referred to the ideal orbit, divided by the reference momentum
    double PYC;//vertical canonical momentum of the (closed) orbit referred to the ideal orbit, divided by the reference momentum
    double S;//position of the element, measured as the distance along the accelerator [m]
    double TC;//velocity of light times the negative time difference with respect to the reference particle [m]
    double XC;//horizontal position x of the (closed) orbit, referred to the ideal orbit [m]
    double YC;//vertical position y of the (closed) orbit, referred to the ideal orbit [m]
    double K0L;//K0*L, where K0 is the normal dipole coefficient and L the magnetic length
    double K0SL;//K0S*L, K0S is the skew dipole coefficient
    double K1L;//K1*L, where K1 is the normal quadrupole coefficient and L the magnetic length
    double K1SL;//K1S*L, K1S is the skew quadrupole coefficient
    double K2L;//K2*L, where K2 is the normal sextupole coefficient and L the magnetic length
    double K2SL;//K2S*L, K2S is the skew sextupole coefficient
    string PARENT;

    double aperx;//aperture in x
    double apery;//aperture in y

};

#endif
