#ifndef FLUKACOLLIMATOR_H
#define FLUKACOLLIMATOR_H

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

#include <unistd.h>
#include <signal.h>
#include <assert.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cmath>
#if defined(FLUKA)
#include "FlukaIO.h"
#endif
#include "Collimator.h"

using namespace std;


/*
#############################################################################################################

This class describes all Fluka collimators in the accelerator lattice.

#############################################################################################################
*/


class FlukaCollimator : public Collimator
{
public:

    //=================================================Constructors, destructor=======================================

    FlukaCollimator(const double& ALFX, const double& ALFY, const double& APER_1, const double& APER_2, const double& APER_3, const double& APER_4, const string& APERTYPE, const double& BETX, const double& BETY, const double& DPX, const double& DPY, const double& DX, const double& DY, const string& KEYWORD, const double& L, const double& MUX, const double& MUY, const string& NAME, const double& PTC, const double& PXC, const double& PYC, const double& S, const double& TC, const double& XC, const double& YC, const double& K0L, const double& K0SL, const double& K1L, const double& K1SL, const double& K2L, const double& K2SL, const string& PARENT, const string& meth, const long double& hgap, const long double& hgap2, const double& collang, const long double& pdepth, const long double& pdepth2, const double& tcang, const double& nsig);

    FlukaCollimator(Element elt, const double& tcang, const double& nsig, const string& meth);

    FlukaCollimator(const FlukaCollimator& obj);

    virtual ~FlukaCollimator() {};


    //method usually describing the passage through a collimator; here it is not used

    void collipass(Particle& p1, double& dpopeff, const double& scaleorbit, const double& R11X, const double& R12X, const double& R21X, const double& R22X, const double& R11Y, const double& R12Y, const double& R21Y, const double& R22Y, const double& dx1, const double& dpx1, const double& dy1, const double& dpy1, const double& delta_s, const double& Apr, const double& Zpr, const double& betgam) {};

    //displays the parameters of the collimator

    void affiche();


#if defined(FLUKA)
    //specific methods describing the passage of a bunch of particles through a Fluka collimator

    void collipassfluka(vector <Particle>& bunchstart, vector <Particle>& bunchend, flukaio_connection_t* connection, const int& turn, const double& momentum);


    //change of referential for the use of Fluka

    void convertToFluka(const int& turn, particle_info_t& part, const Particle& p1, const double& p0);


    //change of referential after the passage through Fluka to return to ICOSIM++

    void convertFromFluka(Particle& p1, const particle_info_t& particle, double& p0);


    //ends a Fluka connection

    void die(char* msg, flukaio_connection_t* conn);
#endif

};

#endif
