#ifndef PARTICLE_H
#define PARTICLE_H
#include <iostream>
#include <vector>
#include <string>
#include "stdlib.h"
#include <math.h>
using namespace std;

/*
#############################################################################################################################################################################################

This class describes the particles in the experiment. It has as attribute the 6-dimensional coordinates and as methods some usefull tools related to the tracking of the particles.

This is also here that we have the method which generates particles.

#############################################################################################################################################################################################
 */


class Particle
{

public:

    //==================Constructors, destructor=========================================

    Particle(double x = 0, double dx = 0, double y = 0, double dy = 0, double deltaE = 0, double t = 0, double massnumber = 208, double chargestate = 82, double moment = 1.2);

    Particle(const Particle& obj);

    virtual ~Particle() {};


    //method which display the 6 coordinates of a particles before and after the passage through an element

    void afficheCoordonnees();


    //displays only the 6 coordinates of the particle before the passage through an element

    void afficheFirstCoordonnees();


    //generates a particle. We can choose between the following distributions: kv, waterbag or r1r2.

    void genpartdist(const int& i0, const int& n, const string& type, const double& r1r2skin, const double& emx, const double& emy, const double& sigdpp, const double& bx, const double& ax, const double& dx, const double& dpx, const double& by, const double& ay, const double& dy, const double& dpy, const double& nsigi);


    //gamma incomplete function; the base source code is taken from  www.crbond.com/math.htm, but has been modified

    int incog(double a, double x, double& gin, double& gim, double& gip);


    //returns a random double in the interval [0,1]. The distribution is uniformed.

    double random();


    //linear interpolation method

    double interp1(double x1, double y1, double x2, double y2, double x);


    //get and set methodes for the identification

    int getidentification() const;

    void setidentification(int id);

    Particle& operator=(Particle const& source);


    double Ap0; //massnumber of particle
    double Zp0; //chargestate of particle
    double dpoporiginal; //momentum (normalement on n en a pas besoin, = p1.coordonnees[][4]
    int inabs; //0 if gets lost, 1 if continue

    long double coordonnees[2][6];
    long double dt;
    int nrevhitp;
    int in;//0 if gets lost, 1 if continue (used only in trackensemblelinearnew)

private:

    int identification;//id of the particle



};


#endif
